package ucl.cities.model;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunState;
import repast.simphony.space.graph.Network;
import repast.simphony.space.graph.RepastEdge;

import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.City;
import ucl.cities.utilities.DistanceMeasure;

/**
 * The main Wilson model implemented using an agent-based method.
 * 
 * See:  Birkin, M. and Heppenstall, A. 2011. Extending spatial 
 *	interaction models with agents for understanding relationships in a
 * dynamic retail market. Urban Studies Research. doi: 10.1155/2011/403969.
 *  
 * @author maltaweel
 *
 */
public class WilsonAgentModel implements Model{

	public static double Constant=1;  //k
//	public static double F;  //f
	public static double Epsilon1=1;  // e1
//	public static double Epsilon2;  // e2
//	public static double Epsilon3;  // e3
	Map<String,RepastEdge>flowEdges = new HashMap<String,RepastEdge>();
	
	/**
	 * The agent version of Wilson's model main constructor.
	 */
	public WilsonAgentModel(){
		
	}
	
	/**
	 * The basic iteration for the Wilson model (1-4).
	 * @param c a city object.
	 */
	public void step(City c){
		flow(c);
	//	calculateSize(c);
	}
	
	/**
	 * This is the flow equation based on Harris, B and Wilson, A.G. 1978.
	 * "Equilibrium values and dynamics of attractiveness terms in 
	 * production-constrained spatial interaction models. 
	 * Environment and Planning A 10:371-388.
	 * @param c a city object.
	 */
	public void flow (City c){
		Context context=RunState.getSafeMasterContext();
		Iterator<City> ci=context.getObjects(City.class).iterator();
		Network flowN = (Network) RunState.getInstance().getMasterContext().getProjection("Flow");
		
		while(ci.hasNext()){
			City c2 = ci.next();
			double distance=0;
			if(c==c2)
				distance=0.001;
			distance=DistanceMeasure.measureDistanceBetweenCities(c, c2)/10000; //for kilometers
			double betaCal=Math.exp(-c.getBeta()*distance);
			double alphaCal=Math.pow(c2.getSize(),c2.externalFlowPercentage*c2.getAlpha());
			
			double flow=c.getScalars()*c.getPopulation()*alphaCal*betaCal;

			String x=c.getName();
			String y=c2.getName();
			
			String name=x+"-"+y;
			
			RepastEdge e=null;
			if(flowEdges.containsKey(name)){
				e=flowEdges.get(name);
				e.setWeight(Math.log10(flow));
			}
			else{
				e = new RepastEdge(c,c2,true,Math.log10(flow));
				context.add(e);
				flowN.addEdge(e);
				flowEdges.put(name,e);
			}
			
			for(CanalEdge ccc: DistanceMeasure.path){
				ccc.setUsed(flow);
			}
			c2.setFlowValue(Math.log10(flow)+(c2.getFlowValue()));
		}
	}
	
	/**
	 * Method to calculate city size based on Harris and Wilson (1978).
	 * @param c a city object.
	 */
	public void calculateSize(City c){
		double revenue=c.getFlowValue();
		double size=c.getSize()+Epsilon1*(revenue-Constant*c.getSize());
		c.setSize(size);
	}

	public Map<String, RepastEdge> getFlowEdges() {
		return flowEdges;
	}

	public void setFlowEdges(Map<String, RepastEdge> flowEdges) {
		this.flowEdges = flowEdges;
	}
	
	/**
	 * Adjust flow based on external flow as well
	 * @param revenue the overall revenue
	 * @param c the city
	 * @return
	 */
//	public double externaFlow(double revenue, City c){
//		return revenue=revenue+(c.externalFlowPercentage*revenue);
//	}
	
	/**
	 * The step into method.
	 * See (6) in article.
	 * 
	 * @param c the city variable.
	 */
/*	public void stepFirstOption(City c){
		if(c.isInteractSize()){
			c.delta=1;
		}
		else
			c.delta=0;
		
		c.s=Math.exp(-c.getBeta()-c.getLambda()*c.getPrice());
	}
*/	
	/**
	 * See step (6) in article.
	 * 
	 * @param c a city variable.
	 */
/*	public void sumSignma(City c){
		Context context=RunState.getSafeMasterContext();
		Iterator<City> ci=context.getObjects(City.class).iterator();
		
		double sigTotal=0;
		while(ci.hasNext()){
			City cc = ci.next();
			if(c==cc)
				continue;
			sigTotal+=cc.getS();
		}
		
		while(ci.hasNext()){
			City cc = ci.next();
			c.s=(c.getS()/sigTotal)*cc.getH()*this.F;
		}
	}
*/	
	/**
	 * Calculate w using (7) in the paper.
	 * 
	 * @param c a city variable.
	 */
/*	public void w(City c){
		double size=c.getSize()*(1+(Epsilon1*(c.getRevenue()-c.getCost())/(c.getCost())));
		c.setSize(size);
	}
*/
	/**
	 * Calculate p using (7) in the paper.
	 * 
	 * @param c a city variable.
	 */
/*	public void p(City c){
		double price=c.getPrice()*(1+(Epsilon2*(c.getRevenue()-c.getCost())/c.getCost()));
		c.setPrice(price);
	}
*/
	/**
	 * Dynamic rent scheme, see (9) in the paper.
	 * @param c
	 */
/*	public void dynamicRent(City c){
		double rent=c.getRent()+Epsilon3*(c.getRevenue()-c.getCost());
		c.setRent(rent);
	} 
*/
	/**
	 * Dynamic cost, see (8) in the paper.
	 * @param c
	 */
/*	public void dynamicCos(City c){
		double cost=Constant*c.getSize()+c.getRent();
		c.setCost(cost);
	}
*/	
	
	
}
