package ucl.cities.utilities;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.PriorityQueue;

import repast.simphony.space.graph.Network;
import ucl.cities.entities.CanalEdge;
import ucl.cities.entities.Location;

/**
 * Returns the shortest path as a list of edges
 * See ShortestPath.class and Wikipedia for further details
 * @author Yanwei Wu
 */

public class DijkstraShortestPath{

	private Network network;
//	private ArrayList<CanalEdge> pathList;
	private ArrayList settled; 
	private PriorityQueue<Location> unsettled;
	private double leastCost;
	
	/**
	 * Primary constructor for object
	 * @param network a network object
	 */
	public DijkstraShortestPath(Network network) {
		this.network = network;
		this.leastCost = 0;
		
	}

	/**
	 * Method returns shortest path between the source and target node
	 * @param source a source node
	 * @param target a target node
	 * @return
	 */
	public ArrayList<CanalEdge> getPath(Object source, Object target){
		ArrayList<CanalEdge> pathList = new ArrayList<CanalEdge>();
		settled = new ArrayList();	
		unsettled = new PriorityQueue<Location>( 11,
				new Comparator<Location>() {
					public int compare(Location o1, Location o2) {
						return o1.getGn() < o2.getGn() ? -1 : o1.getGn() == o2.getGn() ? 0 : 1;
					}
				});
		

		Location sourceVertex = (Location) source;
		Location targetVertex = (Location) target;
		sourceVertex.setGn(0);
		
		unsettled.add(sourceVertex);
		while(!unsettled.isEmpty()){
			Location u = unsettled.poll();
			settled.add(u);
			relax(u);
		}
	
		pathList = getPathList(sourceVertex, targetVertex);
		this.leastCost = targetVertex.getGn();
		
		resetNetwork();
		return pathList;	
	}
	
	public void relax(Location u){
		for(Object edge0: network.getEdges(u)){
			CanalEdge edge = (CanalEdge)(edge0);
			Location v = (Location) edge.getTarget();
			double weight = edge.getWeight();
			if (v.equals(u))
			{
				if (!network.isDirected())
				{
					v = (Location) edge.getSource();
					weight = edge.getWeight();
				}
				else
					continue;
			}
			
			if (settled.contains(v)) continue;
			
			if ((u.getGn() + weight) < v.getGn())
			{
				if(unsettled.contains(v))
					unsettled.remove(v);
				v.setGn(u.getGn() + weight);
				v.setParent(u);
				unsettled.add(v);
			}
		}
	}

	public ArrayList<CanalEdge> getPathList(Location source, Location target)
	{
		ArrayList<CanalEdge> list = new ArrayList<CanalEdge>();

		Location tmp = target;	
		if (tmp.getParent() != null) 
		{
			while(tmp != source)
			{
				CanalEdge edge = (CanalEdge) network.getEdge(tmp.getParent(), tmp);
				double weight = edge.getWeight();
				if (edge.getSource().equals(tmp)&& !network.isDirected())
				{
				//	edge = (CanalEdge) network.getEdge(tmp, tmp.getParent());
					weight = edge.getWeight();
				}
				list.add(0, edge);
				tmp = tmp.getParent();					
			}				
		}
		return list;
	}
	
	public void resetNetwork()
	{
		Iterator<Location> i = network.getNodes().iterator();
		while(i.hasNext())
		{
			Location n = i.next();
			n.setParent(null);
			n.setGn(Double.POSITIVE_INFINITY);
		}
	}

	public double getLeastCost() {
		return leastCost;
	}

	public void setLeastCost(double leastCost) {
		this.leastCost = leastCost;
	}
	
}
